@@ -1,25 +1,27 @@
-mod peer;
+mod ntp_source;
 mod server;
 pub mod subnet;
 
-use ntp_os_clock::DefaultNtpClock;
-use ntp_proto::{SourceDefaultsConfig, SynchronizationConfig};
-use ntp_udp::{EnableTimestamps, InterfaceName};
-pub use peer::*;
+use clock_steering::unix::UnixClock;
+#[cfg(feature = "unstable_ntpv5")]
+use ntp_proto::NtpVersion;
+use ntp_proto::{AlgorithmConfig, SourceDefaultsConfig, SynchronizationConfig};
+pub use ntp_source::*;
 use serde::{Deserialize, Deserializer};
 pub use server::*;
 use std::{
+    fmt::Display,
     io::ErrorKind,
     net::SocketAddr,
     os::unix::fs::PermissionsExt,
     path::{Path, PathBuf},
     str::FromStr,
 };
-use thiserror::Error;
+use timestamped_socket::interface::InterfaceName;
 use tokio::{fs::read_to_string, io};
 use tracing::{info, warn};
 
-use super::tracing::LogLevel;
+use super::{clock::NtpClockWrapper, tracing::LogLevel};
 
 const USAGE_MSG: &str = "\
 usage: ntp-daemon [-c PATH] [-l LOG_LEVEL]
@@ -99,7 +101,7 @@ impl CliArg {
                     // split combined shorthand options
                     for (n, char) in short_arg.trim_start_matches('-').chars().enumerate() {
                         let flag = format!("-{char}");
-                        // convert option argument to seperate segment
+                        // convert option argument to separate segment
                         if takes_argument_short.contains(&char) {
                             let rest = short_arg[(n + 2)..].trim().to_string();
                             // assignment syntax is not accepted for shorthand arguments
@@ -208,7 +210,7 @@ impl NtpDaemonOptions {
     }
 }
 
-fn deserialize_ntp_clock<'de, D>(deserializer: D) -> Result<DefaultNtpClock, D::Error>
+fn deserialize_ntp_clock<'de, D>(deserializer: D) -> Result<NtpClockWrapper, D::Error>
 where
     D: Deserializer<'de>,
 {
@@ -216,10 +218,16 @@ where
 
     if let Some(path) = data {
         tracing::info!("using custom clock {path:?}");
-        DefaultNtpClock::from_path(&path).map_err(|e| serde::de::Error::custom(e.to_string()))
+        #[cfg(target_os = "linux")]
+        return Ok(NtpClockWrapper::new(
+            UnixClock::open(path).map_err(|e| serde::de::Error::custom(e.to_string()))?,
+        ));
+
+        #[cfg(not(target_os = "linux"))]
+        panic!("Custom clock paths not supported on this platform");
     } else {
         tracing::debug!("using REALTIME clock");
-        Ok(DefaultNtpClock::realtime())
+        Ok(NtpClockWrapper::new(UnixClock::CLOCK_REALTIME))
     }
 }
 
@@ -230,22 +238,67 @@ where
     let opt_interface_name: Option<InterfaceName> = Deserialize::deserialize(deserializer)?;
 
     if let Some(interface_name) = opt_interface_name {
-        tracing::info!("using custom interface {}", interface_name);
+        tracing::debug!("using custom interface {}", interface_name);
     } else {
-        tracing::info!("using default interface");
+        tracing::trace!("using default interface");
     }
 
     Ok(opt_interface_name)
 }
 
+/// Timestamping mode. This is a hint!
+///
+/// Your OS or hardware might not actually support some timestamping modes.
+/// Unsupported timestamping modes are ignored.
+#[derive(Default, Debug, Clone, Copy, Deserialize, PartialEq, Eq, Hash)]
+#[serde(rename_all = "kebab-case")]
+pub enum TimestampMode {
+    #[cfg_attr(not(any(target_os = "linux", target_os = "freebsd")), default)]
+    Software,
+    #[cfg_attr(target_os = "freebsd", default)]
+    KernelRecv,
+    #[cfg_attr(target_os = "linux", default)]
+    KernelAll,
+    Hardware,
+}
+
+impl TimestampMode {
+    #[cfg(target_os = "linux")]
+    pub(crate) fn as_interface_mode(self) -> timestamped_socket::socket::InterfaceTimestampMode {
+        use timestamped_socket::socket::InterfaceTimestampMode::*;
+        match self {
+            TimestampMode::Software => None,
+            TimestampMode::KernelRecv => SoftwareRecv,
+            TimestampMode::KernelAll => SoftwareAll,
+            TimestampMode::Hardware => HardwareAll,
+        }
+    }
+
+    #[cfg(any(target_os = "linux", target_os = "freebsd"))]
+    pub(crate) fn as_general_mode(self) -> timestamped_socket::socket::GeneralTimestampMode {
+        use timestamped_socket::socket::GeneralTimestampMode::*;
+        match self {
+            TimestampMode::Software => None,
+            TimestampMode::KernelRecv => SoftwareRecv,
+            TimestampMode::KernelAll | TimestampMode::Hardware => SoftwareAll,
+        }
+    }
+
+    #[cfg(not(any(target_os = "linux", target_os = "freebsd")))]
+    pub(crate) fn as_general_mode(self) -> timestamped_socket::socket::GeneralTimestampMode {
+        use timestamped_socket::socket::GeneralTimestampMode::*;
+        None
+    }
+}
+
 #[derive(Deserialize, Debug, Copy, Clone, Default)]
 #[serde(rename_all = "kebab-case", deny_unknown_fields)]
 pub struct ClockConfig {
     #[serde(deserialize_with = "deserialize_ntp_clock", default)]
-    pub clock: DefaultNtpClock,
+    pub clock: NtpClockWrapper,
     #[serde(deserialize_with = "deserialize_interface", default)]
     pub interface: Option<InterfaceName>,
-    pub enable_timestamps: EnableTimestamps,
+    pub timestamp_mode: TimestampMode,
 }
 
 #[derive(Deserialize, Debug, Clone)]
@@ -253,6 +306,8 @@ pub struct ClockConfig {
 pub struct ObservabilityConfig {
     #[serde(default)]
     pub log_level: Option<LogLevel>,
+    #[serde(default = "default_ansi_colors")]
+    pub ansi_colors: bool,
     #[serde(default)]
     pub observation_path: Option<PathBuf>,
     #[serde(default = "default_observation_permissions")]
@@ -265,6 +320,7 @@ impl Default for ObservabilityConfig {
     fn default() -> Self {
         Self {
             log_level: Default::default(),
+            ansi_colors: default_ansi_colors(),
             observation_path: Default::default(),
             observation_permissions: default_observation_permissions(),
             metrics_exporter_listen: default_metrics_exporter_listen(),
@@ -272,6 +328,10 @@ impl Default for ObservabilityConfig {
     }
 }
 
+const fn default_ansi_colors() -> bool {
+    true
+}
+
 const fn default_observation_permissions() -> u32 {
     0o666
 }
@@ -280,17 +340,27 @@ fn default_metrics_exporter_listen() -> SocketAddr {
     "127.0.0.1:9975".parse().unwrap()
 }
 
+#[derive(Deserialize, Debug, Default)]
+#[serde(rename_all = "kebab-case", deny_unknown_fields)]
+pub struct DaemonSynchronizationConfig {
+    #[serde(flatten)]
+    pub synchronization_base: SynchronizationConfig,
+
+    #[serde(default)]
+    pub algorithm: AlgorithmConfig,
+}
+
 #[derive(Deserialize, Debug, Default)]
 #[serde(rename_all = "kebab-case", deny_unknown_fields)]
 pub struct Config {
     #[serde(rename = "source", default)]
-    pub sources: Vec<PeerConfig>,
+    pub sources: Vec<NtpSourceConfig>,
     #[serde(rename = "server", default)]
     pub servers: Vec<ServerConfig>,
     #[serde(rename = "nts-ke-server", default)]
     pub nts_ke: Vec<NtsKeConfig>,
     #[serde(default)]
-    pub synchronization: SynchronizationConfig,
+    pub synchronization: DaemonSynchronizationConfig,
     #[serde(default)]
     pub source_defaults: SourceDefaultsConfig,
     #[serde(default)]
@@ -329,7 +399,7 @@ impl Config {
             info!("using config file at default location `{:?}`", global_path);
             match Config::from_file(global_path).await {
                 Err(ConfigError::Io(e)) if e.kind() == ErrorKind::PermissionDenied => {
-                    info!("permission denied on global config file! using default config ...");
+                    warn!("permission denied on global config file! using default config ...");
                 }
                 other => {
                     return other;
@@ -342,16 +412,16 @@ impl Config {
 
     pub async fn from_args(
         file: Option<impl AsRef<Path>>,
-        peers: Vec<PeerConfig>,
+        sources: Vec<NtpSourceConfig>,
         servers: Vec<ServerConfig>,
     ) -> Result<Config, ConfigError> {
         let mut config = Config::from_first_file(file.as_ref()).await?;
 
-        if !peers.is_empty() {
+        if !sources.is_empty() {
             if !config.sources.is_empty() {
-                info!("overriding peers from configuration");
+                info!("overriding sources from configuration");
             }
-            config.sources = peers;
+            config.sources = sources;
         }
 
         if !servers.is_empty() {
@@ -364,16 +434,17 @@ impl Config {
         Ok(config)
     }
 
-    /// Count potential number of peers in configuration
-    fn count_peers(&self) -> usize {
+    /// Count potential number of sources in configuration
+    fn count_sources(&self) -> usize {
         let mut count = 0;
-        for peer in &self.sources {
-            match peer {
-                PeerConfig::Standard(_) => count += 1,
-                PeerConfig::Nts(_) => count += 1,
-                PeerConfig::Pool(config) => count += config.max_peers,
+        for source in &self.sources {
+            match source {
+                NtpSourceConfig::Standard(_) => count += 1,
+                NtpSourceConfig::Nts(_) => count += 1,
+                NtpSourceConfig::Pool(config) => count += config.count,
                 #[cfg(feature = "unstable_nts-pool")]
-                PeerConfig::NtsPool(config) => count += config.max_peers,
+                NtpSourceConfig::NtsPool(config) => count += config.count,
+                NtpSourceConfig::Sock(_) => count += 1,
             }
         }
         count
@@ -393,28 +464,64 @@ impl Config {
         }
 
         if !self.sources.is_empty()
-            && self.count_peers() < self.synchronization.minimum_agreeing_sources
+            && self.count_sources()
+                < self
+                    .synchronization
+                    .synchronization_base
+                    .minimum_agreeing_sources
         {
             warn!("Fewer sources configured than are required to agree on the current time. Daemon will not change system time.");
             ok = false;
         }
 
+        #[cfg(feature = "unstable_ntpv5")]
+        if self.sources.iter().any(|config| match config {
+            NtpSourceConfig::Sock(_) => false,
+            NtpSourceConfig::Standard(config) => matches!(config.ntp_version, Some(NtpVersion::V5)),
+            NtpSourceConfig::Nts(config) => matches!(config.ntp_version, Some(NtpVersion::V5)),
+            NtpSourceConfig::Pool(config) => matches!(config.ntp_version, Some(NtpVersion::V5)),
+            #[cfg(feature = "unstable_nts-pool")]
+            NtpSourceConfig::NtsPool(config) => matches!(config.ntp_version, Some(NtpVersion::V5)),
+        }) {
+            warn!("Forcing a source into NTPv5, which is still a draft. There is no guarantee that the server will remain compatible with this or future versions of ntpd-rs.");
+            ok = false;
+        }
+
         ok
     }
 }
 
-#[derive(Error, Debug)]
+#[derive(Debug)]
 pub enum ConfigError {
-    #[error("io error while reading config: {0}")]
-    Io(#[from] io::Error),
-    #[error("config toml parsing error: {0}")]
-    Toml(#[from] toml::de::Error),
+    Io(io::Error),
+    Toml(toml::de::Error),
+}
+
+impl std::error::Error for ConfigError {}
+
+impl Display for ConfigError {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        match self {
+            Self::Io(e) => write!(f, "io error while reading config: {e}"),
+            Self::Toml(e) => write!(f, "config toml parsing error: {e}"),
+        }
+    }
+}
+
+impl From<io::Error> for ConfigError {
+    fn from(value: io::Error) -> Self {
+        Self::Io(value)
+    }
+}
+
+impl From<toml::de::Error> for ConfigError {
+    fn from(value: toml::de::Error) -> Self {
+        Self::Toml(value)
+    }
 }
 
 #[cfg(test)]
 mod tests {
-    use std::str::FromStr;
-
     use ntp_proto::{NtpDuration, StepThreshold};
 
     use super::*;
@@ -425,8 +532,10 @@ mod tests {
             toml::from_str("[[source]]\nmode = \"server\"\naddress = \"example.com\"").unwrap();
         assert_eq!(
             config.sources,
-            vec![PeerConfig::Standard(StandardPeerConfig {
+            vec![NtpSourceConfig::Standard(StandardSource {
                 address: NormalizedAddress::new_unchecked("example.com", 123).into(),
+                #[cfg(feature = "unstable_ntpv5")]
+                ntp_version: None,
             })]
         );
         assert!(config.observability.log_level.is_none());
@@ -438,8 +547,10 @@ mod tests {
         assert_eq!(config.observability.log_level, Some(LogLevel::Info));
         assert_eq!(
             config.sources,
-            vec![PeerConfig::Standard(StandardPeerConfig {
+            vec![NtpSourceConfig::Standard(StandardSource {
                 address: NormalizedAddress::new_unchecked("example.com", 123).into(),
+                #[cfg(feature = "unstable_ntpv5")]
+                ntp_version: None,
             })]
         );
 
@@ -449,16 +560,26 @@ mod tests {
             .unwrap();
         assert_eq!(
             config.sources,
-            vec![PeerConfig::Standard(StandardPeerConfig {
+            vec![NtpSourceConfig::Standard(StandardSource {
                 address: NormalizedAddress::new_unchecked("example.com", 123).into(),
+                #[cfg(feature = "unstable_ntpv5")]
+                ntp_version: None,
             })]
         );
         assert_eq!(
-            config.synchronization.single_step_panic_threshold.forward,
+            config
+                .synchronization
+                .synchronization_base
+                .single_step_panic_threshold
+                .forward,
             Some(NtpDuration::from_seconds(0.))
         );
         assert_eq!(
-            config.synchronization.single_step_panic_threshold.backward,
+            config
+                .synchronization
+                .synchronization_base
+                .single_step_panic_threshold
+                .backward,
             Some(NtpDuration::from_seconds(0.))
         );
 
@@ -468,17 +589,21 @@ mod tests {
             .unwrap();
         assert_eq!(
             config.sources,
-            vec![PeerConfig::Standard(StandardPeerConfig {
+            vec![NtpSourceConfig::Standard(StandardSource {
                 address: NormalizedAddress::new_unchecked("example.com", 123).into(),
+                #[cfg(feature = "unstable_ntpv5")]
+                ntp_version: None,
             })]
         );
         assert!(config
             .synchronization
+            .synchronization_base
             .single_step_panic_threshold
             .forward
             .is_none());
         assert!(config
             .synchronization
+            .synchronization_base
             .single_step_panic_threshold
             .backward
             .is_none());
@@ -508,8 +633,10 @@ mod tests {
 
         assert_eq!(
             config.sources,
-            vec![PeerConfig::Standard(StandardPeerConfig {
+            vec![NtpSourceConfig::Standard(StandardSource {
                 address: NormalizedAddress::new_unchecked("example.com", 123).into(),
+                #[cfg(feature = "unstable_ntpv5")]
+                ntp_version: None,
             })]
         );
 
@@ -563,7 +690,7 @@ mod tests {
     }
 
     #[test]
-    fn toml_peers_invalid() {
+    fn toml_sources_invalid() {
         let config: Result<Config, _> = toml::from_str(
             r#"
             [[source]]
@@ -576,7 +703,7 @@ mod tests {
     }
 
     #[test]
-    fn toml_allow_no_peers() {
+    fn toml_allow_no_sources() {
         let config: Result<Config, _> = toml::from_str(
             r#"
             [[server]]
@@ -684,8 +811,7 @@ mod tests {
         let config: Result<ClockConfig, _> = toml::from_str(
             r#"
             interface = "enp0s31f6"
-            enable-timestamps.rx-hardware = true
-            enable-timestamps.tx-software = true
+            timestamp-mode = "software"
             "#,
         );
 
@@ -694,7 +820,30 @@ mod tests {
         let expected = InterfaceName::from_str("enp0s31f6").unwrap();
         assert_eq!(config.interface, Some(expected));
 
-        assert!(config.enable_timestamps.rx_software);
-        assert!(config.enable_timestamps.tx_software);
+        assert_eq!(config.timestamp_mode, TimestampMode::Software);
+    }
+
+    #[test]
+    fn daemon_synchronization_config() {
+        let config: Result<DaemonSynchronizationConfig, _> = toml::from_str(
+            r#"
+            does_not_exist = 5
+            "#,
+        );
+
+        assert!(config.is_err());
+
+        let config: Result<DaemonSynchronizationConfig, _> = toml::from_str(
+            r#"
+            minimum-agreeing-sources = 2
+
+            [algorithm]
+            initial-wander = 1e-7
+            "#,
+        );
+
+        let config = config.unwrap();
+        assert_eq!(config.synchronization_base.minimum_agreeing_sources, 2);
+        assert_eq!(config.algorithm.initial_wander, 1e-7);
     }
 }
